# -*- coding: utf-8 -*-
from watson.common import strings


def serialize(obj, attributes, strategies=None, camelcase=True, **kwargs):
    """Serializes an object into a dict suitable to be dumped to json.

    Args:
        obj (mixed): The object to serialize
        attributes (tuple): A tuple of attributes that should be serialized
        strategies (dict): Key/value pairs of strategies to deal with objects
        camelcase (bool): camelCase the attribute names
        kwargs: Passed to the strategy when it's called

    Example:

    .. code-block:: python

        class AnotherClass(object):
            name = 'Else'

        class MyClass(object):
            name = 'Something'
            complex_classes = [AnotherClass()]

        class_ = MyClass()
        d = serialize(class_, ('name', 'complex'),
                       strategies={
                            'complex_class': lambda x:
                                                serialize(y, ('name',))
                                                for y in x})
        # {'name': 'Something', 'complexClass': {'name': 'Else'}}
    """
    serialized = {}
    for attr in attributes:
        value = getattr(obj, attr)
        if value is None:
            continue
        if strategies and attr in strategies:
            value = strategies[attr](value, **kwargs)
        if camelcase:
            attr = strings.camelcase(attr, uppercase=False)
        serialized[attr] = value
    return serialized


def deserialize(obj, class_, attributes, strategies=None, snakecase=True,
                **kwargs):
    """Deserializes a dict into an object of type class_.

    Can be seen as the inverse of serialize().

    Args:
        obj (dict): The structure to deserialize
        class_ (class): The type that the object should be deserialized into
        attributes (tuple): A tuple of attributes that should be set
        strategies (dict): Key/value pairs of strategies to deal with objects
        snakecase (bool): snake_case the attribute names
        kwargs: Passed to the strategy when it's called
    """
    deserialized = class_()
    for attr in attributes:
        value = obj.get(attr)
        if value is None:
            continue
        if strategies and attr in strategies:
            value = strategies[attr](value, **kwargs)
        if snakecase:
            attr = strings.snakecase(attr)
        setattr(deserialized, attr, value)
    return deserialized



# -*- coding: utf-8 -*-
from importlib import import_module


definition_lookup = {}


def load_definition_from_string(qualified_module, cache=True):
    """Load a definition based on a fully qualified string.

    Returns:
        None or the loaded object

    Example:

    .. code-block:: python

        definition = load_definition_from_string('watson.http.messages.Request')
        request = definition()
    """
    if qualified_module in definition_lookup and cache:
        return definition_lookup[qualified_module]
    parts = qualified_module.split('.')
    try:
        module = import_module('.'.join(parts[:-1]))
        obj = getattr(module, parts[-1:][0])
        definition_lookup[qualified_module] = obj
        return obj
    except ImportError:
        raise


def get_qualified_name(obj):
    """Retrieve the full module path of an object.

    Example:

    .. code-block:: python

        from watson.http.messages import Request
        request = Request()
        name = get_qualified_name(request) # watson.http.messages.Request
    """
    try:
        name = obj.__qualname__
    except AttributeError:
        try:
            name = obj.__class__.__name__
        except:  # pragma: no cover
            name = obj.__name__  # pragma: no cover
    try:
        module = obj.__module__
        return '{0}.{1}'.format(module, name)
    except:  # pragma: no cover
        return name  # pragma: no cover



# -*- coding: utf-8 -*-
class cached_property(object):

    """Allows expensive property calls to be cached.

    Once the property is called, it's result is stored in the corresponding
    property name prefixed with an underscore.

    Example:

    .. code-block:: python

        class MyClass(object):
            @cached_property
            def expensive_call(self):
                # do something expensive

        klass = MyClass()
        klass.expensive_call  # initial call is made
        klass.expensive_call  # return value is retrieved from an internal cache
        del klass._expensive_call
    """

    def __init__(self, func):
        self.__name__ = func.__name__
        self.__doc__ = func.__doc__
        self.key = '_{name}'.format(name=self.__name__)
        self.func = func

    def __get__(self, obj, type=None):
        if self.key not in obj.__dict__:
            obj.__dict__[self.key] = self.func(obj)
        return obj.__dict__[self.key]



# -*- coding: utf-8 -*-
import re
import unicodedata


def hash(string, replace='_', titlecase=False):
    """Replaces characters with dashes.

    Args:
        replace (string): The character to replace
        titlecase (bool): Titlecase every word
    """
    if titlecase:
        string = string.lower().replace(replace, ' ').title()
        replace = ' '
    return string.replace(replace, '-')


def url_safe(string, separator='-'):
    """Converts a string to a format that is suitable for pretty urls.

    Args:
        separator (string): The separator to use for replaced characters
    """
    string = unicodedata.normalize('NFKD', string)\
                        .encode('ascii', 'ignore').decode('ascii')
    string = re.sub(r'(?i)[^a-z0-9\-_]+', separator, string)
    re_escaped = re.escape(separator)
    string = re.sub(r'{}{{2,}}'.format(re_escaped), separator, string)
    return string.strip(separator).lower()


def camelcase(string, uppercase=True):
    """Converts a string to camelCase.

    Args:
        uppercase (bool): Whether or not to capitalize the first character
    """
    if uppercase:
        return re.sub(r'(?:^|_)(.)', lambda s: s.group(1).upper(), string)
    else:
        return string[0].lower() + camelcase(string)[1:]


def snakecase(string):
    """Converts a string to snake_case.
    """
    string = re.sub(r'([A-Z]+)([A-Z][a-z])', r'\1_\2', string)
    string = re.sub(r'([a-z\d])([A-Z])', r'\1_\2', string)
    return string.replace('-', '_').lower()



# -*- coding: utf-8 -*-
from contextlib import contextmanager

try:
    from contextlib import suppress
except:
    @contextmanager
    def suppress(*exceptions):
        """Provides the ability to not have to write try/catch blocks when just
        passing on the except.

        Thanks to Raymond Hettinger from "Transforming Code into Beautiful
        Idiotmatic Python"
        This will be included in the standard library in 3.4.

        Args:
            exceptions: A list of exceptions to ignore

        Example:

        .. code-block:: python

            # instead of...
            try:
                do_something()
            except:
                pass

            # use this:
            with ignored(Exception):
                do_something()
        """
        try:
            yield
        except exceptions:
            pass

    # Deprecated
    ignored = suppress
